package com.sillycycle.bagleyd.welltris;

/*
 * @(#)PolyominoImage.java
 *
 * createImages() originally written in C
 * Copyright 2000  Stephen Montgomery-Smith
 *
 * Copyright 1993 - 2018  David A. Bagley, bagleyd AT verizon.net
 *
 * Taken from GENERIC X-BASED TETRIS
 *
 * Copyright 1992 - 1995 Q. Alex Zhao, azhao AT cc.gatech.edu
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import android.graphics.Canvas;
import android.graphics.Bitmap;
import android.graphics.Paint;
import android.graphics.Path;
import android.graphics.Point;
import android.graphics.Rect;
import com.sillycycle.bagleyd.util.Colour;

/**
 * The <code>PolyominoImage</code> class holds the colors and
 * handles the construction of the images for the different
 * polyominoes.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/altris.html
 */

public class PolyominoImage {
	//private static final long serialVersionUID = 42L;
	boolean plainPiece = false;
	private static final boolean SMALL_BELLYBUTTON = false;
	private static final boolean LARGE_BELLYBUTTON = false;
	private static final boolean BG_DARK = true;

	static String[] colorStrings = { // 71
		"Red", "Green", "Blue", "Yellow",
		"Magenta", "Cyan", "FireBrick", "GreenYellow",
		"SlateBlue", "Khaki", "Plum", "Violet",
		"DarkTurquoise", "Gold", "Orchid", "Turquoise",
		"Orange", "OrangeRed", "VioletRed", "BlueViolet",
		"SeaGreen", "Pink", "ForestGreen", "SkyBlue",
		"Coral", "Wheat", "Goldenrod", "IndianRed",
		"SpringGreen", "CornflowerBlue", "Thistle", "Aquamarine",
		"CadetBlue", "LightSteelBlue", "NavyBlue", "SteelBlue",
		"YellowGreen", "DarkViolet", "MediumSeaGreen", "DarkSlateGray",
		"White", "MediumVioletRed", "Sienna", "MediumAquamarine",
		"MediumBlue", "Navy", "DarkOliveGreen", "DarkGreen",
		"DimGray", "Tan", "MediumTurquoise", "DarkSlateBlue",
		"Maroon", "Gray", "Black"
	};

	static Colour[] colorThings = {
		new Colour(Colour.red), new Colour(Colour.green),
			new Colour(Colour.blue), new Colour(Colour.yellow),
		new Colour(Colour.magenta), new Colour(Colour.cyan),
			new Colour(178, 34, 34), new Colour(173, 255, 47),
		new Colour(131, 111, 255), new Colour(255, 246, 143),
			new Colour(255, 187, 255), new Colour(238, 130, 238),
		new Colour(0, 206, 209), new Colour(255, 215, 0),
			new Colour(218, 112, 214), new Colour(0, 229, 238),
		new Colour(255, 165, 0), new Colour(255, 69, 0),
			new Colour(208, 32, 144), new Colour(138, 43, 226),
		new Colour(78, 238, 148), new Colour(238, 169, 184),
			new Colour(34, 139, 34), new Colour(126, 192, 238),
		new Colour(255, 127, 80), new Colour(245, 222, 179),
			new Colour(218, 165, 32), new Colour(205, 92, 92),
		new Colour(0, 238, 118), new Colour(100, 149, 237),
			new Colour(216, 191, 216), new Colour(127, 255, 212),
		new Colour(95, 158, 160), new Colour(176, 196, 222),
			new Colour(0, 0, 128), new Colour(70, 130, 180),
		new Colour(154, 205, 50), new Colour(148, 0, 211),
			new Colour(60, 179, 113), new Colour(151, 255, 255),
		new Colour(Colour.white), new Colour(199, 21, 133),
			new Colour(238, 121, 66), new Colour(102, 205, 170),
		new Colour(0, 0, 208), new Colour(0, 0, 177),
			new Colour(85, 107, 47), new Colour(0, 100, 0),
		new Colour(170, 170, 170), new Colour(210, 180, 140),
			new Colour(72, 209, 204), new Colour(72, 61, 139),
		new Colour(176, 48, 96), new Colour(Colour.gray50),
			new Colour(Colour.black)
	};

	public static final int BIG = 256;
	public static final int SMALL = 16;
	public static final int NUM_BITMAPS = BIG;
	static final int LEFT = 1; /* (1<<0) */
	static final int DOWN = 2; /* (1<<1) */
	static final int RIGHT = 4; /* (1<<2) */
	static final int UP = 8; /* (1<<3) */
	static final int LEFT_DOWN = 16; /* (1<<4) */
	static final int RIGHT_DOWN = 32; /* (1<<5) */
	static final int RIGHT_UP = 64; /* (1<<6) */
	static final int LEFT_UP = 128; /* (1<<7) */
	static final int DRAW_COLOR = 0;
	static final int XOR_COLOR = -1;
	static final int XOR_CLEAR_COLOR = -2;
	static final int CLEAR_COLOR = -3;
	StringBuilder[] data = null;
	private static final char[] PATTERN_CHARS = {'*', '.'};

	static boolean IS_LEFT(int n) {
		return ((n & LEFT) != 0);
	}

	static boolean IS_DOWN(int n) {
		return ((n & DOWN) != 0);
	}

	static boolean IS_RIGHT(int n) {
		return ((n & RIGHT) != 0);
	}

	static boolean IS_UP(int n) {
		return ((n & UP) != 0);
	}

	static boolean IS_LEFT_DOWN(int n) {
		return ((n & LEFT_DOWN) != 0);
	}

	static boolean IS_RIGHT_DOWN(int n) {
		return ((n & RIGHT_DOWN) != 0);
	}

	static boolean IS_RIGHT_UP(int n) {
		return ((n & RIGHT_UP) != 0);
	}

	static boolean IS_LEFT_UP(int n) {
		return ((n & LEFT_UP) != 0);
	}

	public static int CHECKLEFT(int x) {
		return (x | LEFT);
	}

	public static int CHECKDOWN(int x) {
		return (x | DOWN);
	}

	public static int CHECKRIGHT(int x) {
		return (x | RIGHT);
	}

	public static int CHECKUP(int x) {
		return (x | UP);
	}

	/* Defines to access the bitmaps. */

	void SETBIT(int x, int y) {
		data[y].setCharAt(x, PATTERN_CHARS[1]);
	}

	void RESBIT(int x, int y) {
		data[y].setCharAt(x, PATTERN_CHARS[0]);
	}

	void HALFBIT(int x, int y) {
		if ((x + y) % 2 == 0)
			RESBIT(x, y);
		else
			SETBIT(x, y);
	}

	void NOTHALFBIT(int x, int y) {
		if ((x + y) % 2 == 0)
			SETBIT(x, y);
		else
			RESBIT(x, y);
	}

	void THIRDBIT(int x, int y) {
		if ((x + y + 2) % 3 == 0)
			SETBIT(x, y);
		else
			RESBIT(x, y);
	}

	void TWOTHIRDSBIT(int x, int y) {
		if ((x + y + 2) % 3 == 0)
			RESBIT(x, y);
		else
			SETBIT(x, y);
	}

	void THREEQUARTERSBIT(int x, int y) {
		if (((y & 1) == 1) || (((x + 2 + y / 2 + 1) & 1) == 1))
			SETBIT(x, y);
		else
			RESBIT(x, y);
	}

	/* Parameters for bitmaps. */
	static int G(int boxsize) { // 1/2 of gap between pentominoes.
		return (boxsize / 45) + 1;
	}

	static int T(int boxsize) { // Thickness of walls of pentominoes.
		return ((boxsize <= 12) ? 1 : G(boxsize * 2));
	}

	static int R(int boxsize) { // Amount of rounding.
		return ((boxsize <= 12) ? 1 : G(boxsize * 6));
	}

	static int RT(int boxsize) { // Thickness of wall of rounded parts.
			// Here 3 is an approximation to 2 sqrt(2).
		return ((boxsize <= 12) ? 1 : G(boxsize * 3));
	}

	static int RR() { // Roof ridge thickness
		return 0;
	}

	/* A list of those bitmaps we need to create to display any polyomino. */
	static int[] imagesNeeded = {
		// This needed for mononimo
		LEFT | RIGHT | UP | DOWN | LEFT_UP | LEFT_DOWN | RIGHT_UP | RIGHT_DOWN,

		LEFT_UP | LEFT_DOWN | RIGHT_UP | RIGHT_DOWN,

		LEFT | RIGHT_UP | RIGHT_DOWN,
		RIGHT | LEFT_UP | LEFT_DOWN,
		UP | LEFT_DOWN | RIGHT_DOWN,
		DOWN | LEFT_UP | RIGHT_UP,
		LEFT | RIGHT_UP,
		RIGHT | LEFT_UP,
		UP | LEFT_DOWN,
		DOWN | LEFT_UP,
		LEFT | RIGHT_DOWN,
		RIGHT | LEFT_DOWN,
		UP | RIGHT_DOWN,
		DOWN | RIGHT_UP,

		// These needed for hexonimoes
		LEFT,
		RIGHT,
		UP,
		DOWN,
		LEFT_DOWN | RIGHT_UP | RIGHT_DOWN,
		LEFT_UP | RIGHT_UP | RIGHT_DOWN,
		LEFT_UP | LEFT_DOWN | RIGHT_DOWN,
		LEFT_UP | LEFT_DOWN | RIGHT_UP,

		LEFT | UP | RIGHT_DOWN,
		LEFT | DOWN | RIGHT_UP,
		RIGHT | UP | LEFT_DOWN,
		RIGHT | DOWN | LEFT_UP,
		LEFT | UP,
		LEFT | DOWN,
		RIGHT | UP,
		RIGHT | DOWN,

		LEFT | RIGHT,
		UP | DOWN,

		RIGHT | UP | DOWN,
		LEFT | UP | DOWN,
		LEFT | RIGHT | DOWN,
		LEFT | RIGHT | UP,

		-1
	};

	static Bitmap[][][] pms = new Bitmap[2][2][NUM_BITMAPS];

	static boolean imageNeeded(int n) {
		for (int i = 0; imagesNeeded[i] != -1; i++)
			if (n == imagesNeeded[i])
				return true;
		return false;
	}

	// http://stackoverflow.com/questions/18630080/android-byte-array-to-bitmap-how-to
	static public Bitmap simpleBitmap(int width, int height,
			Colour color) {
		int size = width * height;
		final int intAlpha = 255;
		int[] intColors = new int[size];

		for (int i = 0; i < size; i++) {
			int modI = i % width;
			int modJ = i / width;
			if ((modI + modJ) % 2 == 0)
				intColors[i] = (intAlpha << 24)
					| (color.getRed() << 16)
					| (color.getGreen() << 8)
					| color.getBlue();
			else
				intColors[i] = (intAlpha << 24)
					| (intAlpha << 16)
					| (intAlpha << 8)
					| intAlpha;
		}
		return Bitmap.createBitmap(intColors, width, height,
			Bitmap.Config.ARGB_8888);
	}

	Bitmap createBitmapPiece(int n, boolean use3D, int boxsize,
			Colour background) {
		Colour[] patternColors = {background, new Colour(0, true)};
		int x, y; // unsigned
		char[] empty = new char[boxsize];

		for (x = 0; x < boxsize; x++)
			empty[x] = '.';
		data = new StringBuilder[boxsize];

		for (y = 0; y < boxsize; y++)
			data[y] = new StringBuilder(new String(empty));
		if (data == null) {
			System.out.println("No images");
			return null; // no biggie, will not use images
		}
		for (y = 0; y < boxsize; y++)
		  for (x = 0; x < boxsize; x++) {
		    if (!use3D) {
		      if (SMALL_BELLYBUTTON) {
			if (x >= boxsize / 2 && x <= boxsize / 2 + 1 &&
			    y >= boxsize / 2 && y <= boxsize / 2 + 1)
				NOTHALFBIT(x, y);
			else
				HALFBIT(x, y);
		      } else
			HALFBIT(x, y);
		    } else if ((x >= y && x <= boxsize - y - 1 && IS_UP(n)) ||
			(x <= y && x <= boxsize - y - 1 && y < boxsize / 2 &&
			  !IS_LEFT(n)) ||
			(x >= y && x >= boxsize - y - 1 && y < boxsize / 2 &&
			  !IS_RIGHT(n))) {
			if (BG_DARK)
				SETBIT(x, y);
			else
				THIRDBIT(x, y);
		    } else if ((x <= y && x <= boxsize - y - 1 && IS_LEFT(n)) ||
			(x >= y && x <= boxsize - y - 1 && x < boxsize / 2 &&
			  !IS_UP(n)) ||
			(x <= y && x >= boxsize - y - 1 && x < boxsize / 2 &&
			  !IS_DOWN(n))) {
			if (BG_DARK)
				TWOTHIRDSBIT(x, y);
			else
				HALFBIT(x, y);
		    } else if ((x >= y && x >= boxsize - y - 1 && IS_RIGHT(n)) ||
			(x >= y && x <= boxsize - y - 1 && x >= boxsize / 2 &&
			  !IS_UP(n)) ||
			(x <= y && x >= boxsize - y - 1 && x >= boxsize / 2 &&
			  !IS_DOWN(n))) {
			if (BG_DARK)
				HALFBIT(x, y);
			else
				TWOTHIRDSBIT(x, y);
		    } else if ((x <= y && x >= boxsize - y - 1 && IS_DOWN(n)) ||
			(x <= y && x <= boxsize - y - 1 && y >= boxsize / 2 &&
			  !IS_LEFT(n)) ||
			(x >= y && x >= boxsize - y - 1 && y >= boxsize / 2 &&
			  !IS_RIGHT(n))) {
			if (BG_DARK)
				THIRDBIT(x, y);
			else
				SETBIT(x, y);
		    }
		  }

		if (IS_LEFT(n))
		  for (y=0;y<boxsize;y++)
		    for (x=G(boxsize);x<G(boxsize)+T(boxsize);x++)
		      SETBIT(x, y);
		if (IS_RIGHT(n))
		  for (y=0;y<boxsize;y++)
		    for (x=G(boxsize);x<G(boxsize)+T(boxsize);x++)
		      SETBIT(boxsize-1-x, y);
		if (IS_UP(n))
		  for (x=0;x<boxsize;x++)
		    for (y=G(boxsize);y<G(boxsize)+T(boxsize);y++)
		      SETBIT(x, y);
		if (IS_DOWN(n))
		  for (x=0;x<boxsize;x++)
		    for (y=G(boxsize);y<G(boxsize)+T(boxsize);y++)
		      SETBIT(x,boxsize-1-y);
		if (IS_LEFT(n))
		  for (y=0;y<boxsize;y++)
		    for (x=0;x<G(boxsize);x++)
		      RESBIT(x, y);
		if (IS_RIGHT(n))
		  for (y=0;y<boxsize;y++)
		    for (x=0;x<G(boxsize);x++)
		      RESBIT(boxsize-1-x, y);
		if (IS_UP(n))
		  for (x=0;x<boxsize;x++)
		    for (y=0;y<G(boxsize);y++)
		      RESBIT(x, y);
		if (IS_DOWN(n))
		  for (x=0;x<boxsize;x++)
		    for (y=0;y<G(boxsize);y++)
		      RESBIT(x,boxsize-1-y);

		if (IS_LEFT(n) && IS_UP(n))
		  for (x=G(boxsize);x<=G(boxsize)+R(boxsize);x++)
		    for (y=G(boxsize);y<=R(boxsize)+2*G(boxsize)-x;y++) {
		      if (x+y>R(boxsize)+2*G(boxsize)-RT(boxsize))
			SETBIT(x, y);
		      else
			RESBIT(x, y);
		    }
		if (IS_LEFT(n) && IS_DOWN(n))
		  for (x=G(boxsize);x<=G(boxsize)+R(boxsize);x++)
		    for (y=G(boxsize);y<=R(boxsize)+2*G(boxsize)-x;y++) {
		      if (x+y>R(boxsize)+2*G(boxsize)-RT(boxsize))
			SETBIT(x,boxsize-1-y);
		      else
			RESBIT(x,boxsize-1-y);
		    }
		if (IS_RIGHT(n) && IS_UP(n))
		  for (x=G(boxsize);x<=G(boxsize)+R(boxsize);x++)
		    for (y=G(boxsize);y<=R(boxsize)+2*G(boxsize)-x;y++) {
		      if (x+y>R(boxsize)+2*G(boxsize)-RT(boxsize))
			SETBIT(boxsize-1-x, y);
		      else
			RESBIT(boxsize-1-x, y);
		    }
		if (IS_RIGHT(n) && IS_DOWN(n))
		  for (x=G(boxsize);x<=G(boxsize)+R(boxsize);x++)
		    for (y=G(boxsize);y<=R(boxsize)+2*G(boxsize)-x;y++) {
		      if (x+y>R(boxsize)+2*G(boxsize)-RT(boxsize))
			SETBIT(boxsize-1-x,boxsize-1-y);
		      else
			RESBIT(boxsize-1-x,boxsize-1-y);
		    }

		if (!IS_LEFT(n) && !IS_UP(n) && IS_LEFT_UP(n)) {
		  for (x=0;x<G(boxsize);x++)
		    for (y=0;y<G(boxsize);y++)
		      RESBIT(x, y);
		  for (x=G(boxsize);x<G(boxsize)+T(boxsize);x++)
		    for (y=0;y<G(boxsize);y++)
		      SETBIT(x, y);
		  for (x=0;x<G(boxsize)+T(boxsize);x++)
		    for (y=G(boxsize);y<G(boxsize)+T(boxsize);y++)
		      SETBIT(x, y);
		}
		if (!IS_LEFT(n) && !IS_DOWN(n) && IS_LEFT_DOWN(n)) {
		  for (x=0;x<G(boxsize);x++)
		    for (y=0;y<G(boxsize);y++)
		      RESBIT(x,boxsize-1-y);
		  for (x=G(boxsize);x<G(boxsize)+T(boxsize);x++)
		    for (y=0;y<G(boxsize);y++)
		      SETBIT(x,boxsize-1-y);
		  for (x=0;x<G(boxsize)+T(boxsize);x++)
		    for (y=G(boxsize);y<G(boxsize)+T(boxsize);y++)
		      SETBIT(x,boxsize-1-y);
		}
		if (!IS_RIGHT(n) && !IS_UP(n) && IS_RIGHT_UP(n)) {
		  for (x=0;x<G(boxsize);x++)
		    for (y=0;y<G(boxsize);y++)
		      RESBIT(boxsize-1-x, y);
		  for (x=G(boxsize);x<G(boxsize)+T(boxsize);x++)
		    for (y=0;y<G(boxsize);y++)
		      SETBIT(boxsize-1-x, y);
		  for (x=0;x<G(boxsize)+T(boxsize);x++)
		    for (y=G(boxsize);y<G(boxsize)+T(boxsize);y++)
		      SETBIT(boxsize-1-x, y);
		}
		if (!IS_RIGHT(n) && !IS_DOWN(n) && IS_RIGHT_DOWN(n)) {
		  for (x=0;x<G(boxsize);x++)
		    for (y=0;y<G(boxsize);y++)
		      RESBIT(boxsize-1-x,boxsize-1-y);
		  for (x=G(boxsize);x<G(boxsize)+T(boxsize);x++)
		    for (y=0;y<G(boxsize);y++)
		      SETBIT(boxsize-1-x,boxsize-1-y);
		  for (x=0;x<G(boxsize)+T(boxsize);x++)
		    for (y=G(boxsize);y<G(boxsize)+T(boxsize);y++)
		      SETBIT(boxsize-1-x,boxsize-1-y);
		}

		if (use3D) {
			if (NUM_BITMAPS != SMALL) {
				if (!IS_LEFT(n) && !IS_UP(n) && !IS_LEFT_UP(n))
				  for (x = 0; x < boxsize / 2 - RR(); x++)
				    for (y = 0; y < boxsize / 2 - RR(); y++)
				      THREEQUARTERSBIT(x, y);
				if (!IS_LEFT(n) && !IS_DOWN(n) && !IS_LEFT_DOWN(n))
				  for (x = 0; x < boxsize / 2 - RR(); x++)
				    for (y = boxsize / 2 + RR(); y < boxsize; y++)
				      THREEQUARTERSBIT(x, y);
				if (!IS_RIGHT(n) && !IS_UP(n) && !IS_RIGHT_UP(n))
				  for (x = boxsize / 2 + RR(); x < boxsize; x++)
				    for (y = 0; y < boxsize / 2 - RR(); y++)
				      THREEQUARTERSBIT(x, y);
				if (!IS_RIGHT(n) && !IS_DOWN(n) && !IS_RIGHT_DOWN(n))
				  for (x = boxsize / 2 + RR(); x < boxsize; x++)
				    for (y = boxsize / 2 + RR(); y < boxsize; y++)
				      THREEQUARTERSBIT(x, y);
			}
		} else if (LARGE_BELLYBUTTON) {
			if (!IS_LEFT(n) && !IS_UP(n) && !IS_LEFT_UP(n))
			  for (x=0;x<G(boxsize)+T(boxsize);x++)
			    for (y=0;y<G(boxsize)+T(boxsize);y++)
			      SETBIT(x, y);
			if (!IS_LEFT(n) && !IS_DOWN(n) && !IS_LEFT_DOWN(n))
			  for (x=0;x<G(boxsize)+T(boxsize);x++)
			    for (y=boxsize-G(boxsize)-T(boxsize);y<boxsize;y++)
			      SETBIT(x, y);
			if (!IS_RIGHT(n) && !IS_UP(n) && !IS_RIGHT_UP(n))
			  for (x=boxsize-G(boxsize)-T(boxsize);x<boxsize;x++)
			    for (y=0;y<G(boxsize)+T(boxsize);y++)
			      SETBIT(x, y);
			if (!IS_RIGHT(n) && !IS_DOWN(n) && !IS_RIGHT_DOWN(n))
			  for (x=boxsize-G(boxsize)-T(boxsize);x<boxsize;x++)
			    for (y=boxsize-G(boxsize)-T(boxsize);y<boxsize;y++)
			      SETBIT(x, y);
		}

		//Pattern
		//return createBitmap(this, true,
		//	stringData, patternChars, patternColors);
		int[] intColors = new int[boxsize * boxsize];
		for (y = 0; y < boxsize; y++)
			for (x = 0; x < boxsize; x++) {
				intColors[x + y * boxsize] =
					((data[y].charAt(x) == '.') ?
					patternColors[1] : patternColors[0]).getColor();
			}
		return Bitmap.createBitmap(intColors, boxsize, boxsize,
			Bitmap.Config.ARGB_8888);
	}

	void createBitmaps(boolean use3D, boolean drawNext,
			int boxsize, Colour background) {
		int n = NUM_BITMAPS - 1;
		int plain = (use3D) ? 0 : 1;
		int next = (drawNext) ? 1 : 0;

		plainPiece = !use3D;
		if ((pms[plain][next][n] = createBitmapPiece(n, use3D,
			boxsize, background)) == null) {
			System.out.println("null: " + n);
			return;
		}
		for (n = 0; n < NUM_BITMAPS - 1; n++) {
			/* Avoid duplication of identical images. */
			if (IS_UP(n) && IS_DOWN(n) && IS_LEFT(n) && IS_RIGHT(n))
				pms[plain][next][n] = pms[plain][next][NUM_BITMAPS - 1];
			else if (IS_LEFT_UP(n) && (IS_LEFT(n) || IS_UP(n)))
				pms[plain][next][n] = pms[plain][next][n & ~LEFT_UP];
			else if (IS_LEFT_DOWN(n) && (IS_LEFT(n) || IS_DOWN(n)))
				pms[plain][next][n] = pms[plain][next][n & ~LEFT_DOWN];
			else if (IS_RIGHT_UP(n) && (IS_RIGHT(n) || IS_UP(n)))
				pms[plain][next][n] = pms[plain][next][n & ~RIGHT_UP];
			else if (IS_RIGHT_DOWN(n) && (IS_RIGHT(n) || IS_DOWN(n)))
				pms[plain][next][n] = pms[plain][next][n & ~RIGHT_DOWN];
			else if (imageNeeded(n) &&
					((pms[plain][next][n] = createBitmapPiece(n, use3D,
					boxsize, background)) == null)) {
				System.out.println("image needed but null: " + n);
				return;
			}
		}
	}

	static int ARR(int i, int j, int[][] shape, int size) {
		return ((i < 0 || i >= size || j < 0 || j >= size)
			? -2 : ((shape[j][i] > 0) ? 1 : 0));
	}

	static int polyPixmap(int i, int j, int[][] shape, int size) {
		int pmid = 0;

		if (ARR(i, j, shape, size) != ARR(i - 1, j, shape, size))
			pmid |= LEFT;
		if (ARR(i, j, shape, size) != ARR(i, j + 1, shape, size))
			pmid |= DOWN;
		if (ARR(i, j, shape, size) != ARR(i + 1, j, shape, size))
			pmid |= RIGHT;
		if (ARR(i, j, shape, size) != ARR(i, j - 1, shape, size))
			pmid |= UP;
		if (NUM_BITMAPS != SMALL) {
			if (ARR(i, j, shape, size) != ARR(i - 1, j + 1, shape, size))
				pmid |= LEFT_DOWN;
			if (ARR(i, j, shape, size) != ARR(i + 1, j + 1, shape, size))
				pmid |= RIGHT_DOWN;
			if (ARR(i, j, shape, size) != ARR(i + 1, j - 1, shape, size))
				pmid |= RIGHT_UP;
			if (ARR(i, j, shape, size) != ARR(i - 1, j - 1, shape, size))
				pmid |= LEFT_UP;
		}
		return pmid;
	}

	public static void drawPoly(Canvas canvas, Paint paint, Point[] points) {
		Path polyPath = new Path();
		polyPath.moveTo(points[0].x, points[0].y);
		for (Point point : points) {
			polyPath.lineTo(point.x, point.y);
		}
		polyPath.lineTo(points[0].x, points[0].y);
		canvas.drawPath(polyPath, paint);
	}

	public static void drawPolygon(Canvas canvas, int lineColor,
			Point[] list) {
		Paint paint = new Paint();
		/*int[] tempListX = new int[n];
		int[] tempListY = new int[n];

		for (int i = 0; i < n; i++) {
			tempListX[i] = list[i].x;
			tempListY[i] = list[i].y;
		}*/
		paint.setColor(lineColor);
		paint.setStyle(Paint.Style.STROKE);
		//paint.setStrokeWidth(5);
		drawPoly(canvas, paint, list);
	}

	public static void fillPolygon(Canvas canvas, int fillColor, int lineColor,
			Point[] list) {
		Paint paint = new Paint();
		/*int[] tempListX = new int[n];
		int[] tempListY = new int[n];

		for (int i = 0; i < n; i++) {
			tempListX[i] = list[i].x;
			tempListY[i] = list[i].y;
		}*/
		paint.setColor(fillColor);
		paint.setStyle(Paint.Style.FILL);
		drawPoly(canvas, paint, list);
		paint.setColor(lineColor);
		paint.setStyle(Paint.Style.STROKE);
		drawPoly(canvas, paint, list);
	}

	static void fillPolygon(Canvas canvas, int cid, int bid,
			int startX, int startY, Point[] list) {
		Point[] tempList = new Point[list.length];

		for (int i = 0; i < list.length; i++) {
			tempList[i] = new Point();
			tempList[i].x = list[i].x + startX;
			tempList[i].y = list[i].y + startY;
		}
		fillPolygon(canvas, cid, bid, tempList);
	}

	static void drawPolygon(Canvas canvas, int cid,
			int startX, int startY, Point[] list) {
		Point[] tempList = new Point[list.length];

		for (int i = 0; i < list.length; i++) {
			tempList[i] = new Point();
			tempList[i].x = list[i].x + startX;
			tempList[i].y = list[i].y + startY;
		}
		drawPolygon(canvas, cid, tempList);
	}

	/* Pure colors can get brighter */
	static int brighter(int color) {
		int alphaPart = (0xFF000000 & color) >> 24;
		int redPart = (0x00FF0000 & color) >> 16;
		int greenPart = (0x0000FF00 & color) >> 8;
		int bluePart = 0x000000FF & color;
		final double FACTOR = 0.8;

		/* From 2D group:
		 * 1. black.brighter() should return grey
		 * 2. applying brighter to blue will always return blue, brighter
		 * 3. non pure color (non zero rgb) will eventually return white		 */
		int i = (int) ((1 - FACTOR) * 255);

		if (redPart >= 0 && redPart < i)
			redPart = i;
		if (greenPart >= 0 && greenPart < i)
			greenPart = i;
		if (bluePart >= 0 && bluePart < i)
			bluePart = i;
		return ((alphaPart << 24) + (redPart << 16)
				+ (greenPart << 8) + bluePart);
	}

	static int darker(int color) {
		int alphaPart = (0xFF000000 & color) >> 24;
		int redPart = (0x00FF0000 & color) >> 16;
		int greenPart = (0x0000FF00 & color) >> 8;
		int bluePart = 0x000000FF & color;
		final double FACTOR = 0.8;

		redPart = (int) (redPart * FACTOR);
		greenPart = (int) (greenPart * FACTOR);
		bluePart = (int) (bluePart * FACTOR);
		return ((alphaPart << 24) + (redPart << 16)
				+ (greenPart << 8) + bluePart);
	}

	void drawWallUnit(Canvas canvas,
			int pmid, int cid, int bid, int x, int y,
			int ulx, int uly, int base_x, int base_y,
			int lrx, int lry, int plateau_x, int plateau_y,
			int draw) {
		int MAX_SIDES = 4;
		Point[] trapazoidList = new Point[MAX_SIDES];

		/* if (color != background && color != xorColor) {
			XSetFillStyle(display, color, FillOpaqueStippled);
			XSetStipple(display, color, graypix);
		}*/
		for (int k = 0; k < MAX_SIDES; k++) {
			trapazoidList[k] = new Point();
		}
		trapazoidList[0].x = x;
		trapazoidList[0].y = y;
		trapazoidList[1].x = base_x;
		trapazoidList[1].y = base_y;
		trapazoidList[2].x = lrx - base_x - ulx;
		trapazoidList[2].y = lry - base_y - uly;
		trapazoidList[3].x = -plateau_x;
		trapazoidList[3].y = -plateau_y;
		/*trapazoidList[4].x = ulx + plateau_x - lrx;
		trapazoidList[4].y = ulx + plateau_y - lry;*/
		for (int k = 0; k < MAX_SIDES; k++) {
			trapazoidList[k].x +=
				((k > 0) ? trapazoidList[k - 1].x : ulx);
			trapazoidList[k].y +=
				((k > 0) ? trapazoidList[k - 1].y : uly);
		}
		fillPolygon(canvas, cid, bid, trapazoidList);
		/* Code added to use pmid info */
		if ((draw == DRAW_COLOR || draw == XOR_COLOR) &&
				!plainPiece) {
			Point[] edgePoints = new Point[MAX_SIDES];
			Point[] pts = new Point[MAX_SIDES];
			int midPointX = 0, midPointY = 0;
			int[] shade = new int[MAX_SIDES];

			if (draw == XOR_COLOR) {
				if (pmid > NUM_BITMAPS) {
					shade[0] = brighter(cid);
					shade[3] = brighter(brighter(cid));
					shade[2] = darker(cid);
					shade[1] = darker(darker(cid));
				} else {
					shade[2] = brighter(cid);
					shade[3] = brighter(brighter(cid));
					shade[0] = darker(cid);
					shade[1] = darker(darker(cid));
				}
			} else {
				if (pmid > NUM_BITMAPS) {
					shade[2] = brighter(cid);
					shade[1] = brighter(brighter(cid));
					shade[0] = darker(cid);
					shade[3] = darker(darker(cid));
				} else {
					shade[0] = brighter(cid);
					shade[1] = brighter(brighter(cid));
					shade[2] = darker(cid);
					shade[3] = darker(darker(cid));
				}
			}
			for (int i = 0; i < MAX_SIDES; i++) {
				midPointX += trapazoidList[i].x;
				midPointY += trapazoidList[i].y;
			}
			midPointX /= MAX_SIDES;
			midPointY /= MAX_SIDES;
			for (int i = 0; i < MAX_SIDES; i++) {
				edgePoints[i] = new Point();
				pts[i] = new Point();
			}

			for (int i = 0; i < MAX_SIDES - 1; i++) {
				edgePoints[i].x = (trapazoidList[i + 1].x +
					trapazoidList[i].x) / 2;
				edgePoints[i].y = (trapazoidList[i + 1].y +
					trapazoidList[i].y) / 2;
			}
			edgePoints[MAX_SIDES - 1].x = (trapazoidList[0].x +
				trapazoidList[MAX_SIDES - 1].x) / 2;
			edgePoints[MAX_SIDES - 1].y = (trapazoidList[0].y +
				trapazoidList[MAX_SIDES - 1].y) / 2;
			for (int i = 0; i < MAX_SIDES; i++) {
				int invi = (MAX_SIDES - i) % MAX_SIDES;
				int invi1 = (MAX_SIDES - i + 3) % MAX_SIDES;
				int invi3 = (MAX_SIDES - i + 1) % MAX_SIDES;
				int i1 = (i + 1) % MAX_SIDES;
				int i3 = (i + 3) % MAX_SIDES;

				if ((pmid & (1 << invi)) == 0) {
					if ((pmid & (1 << invi3)) != 0 ||
						(pmid & (16 << invi)) != 0) {
					pts[0].x = trapazoidList[i3].x;
					pts[0].y = trapazoidList[i3].y;
					pts[1].x = midPointX;
					pts[1].y = midPointY;
					pts[2].x = edgePoints[i3].x;
					pts[2].y = edgePoints[i3].y;
					pts[3].x = trapazoidList[i3].x;
					pts[3].y = trapazoidList[i3].y;
					fillPolygon(canvas, shade[i3], shade[i3], pts);
					}
					if ((pmid & (1 << invi1)) != 0 ||
						(pmid & (16 << invi1)) != 0) {
					pts[0].x = trapazoidList[i].x;
					pts[0].y = trapazoidList[i].y;
					pts[1].x = midPointX;
					pts[1].y = midPointY;
					pts[2].x = edgePoints[i3].x;
					pts[2].y = edgePoints[i3].y;
					pts[3].x = trapazoidList[i].x;
					pts[3].y = trapazoidList[i].y;
					fillPolygon(canvas, shade[i1], shade[i1], pts);
					}
				} else {
					pts[0].x = trapazoidList[i3].x;
					pts[0].y = trapazoidList[i3].y;
					pts[1].x = trapazoidList[i].x;
					pts[1].y = trapazoidList[i].y;
					pts[2].x = midPointX;
					pts[2].y = midPointY;
					pts[3].x = trapazoidList[i3].x;
					pts[3].y = trapazoidList[i3].y;
					fillPolygon(canvas, shade[i], shade[i], pts);
				}
			}
		}
	}

	static void drawWallUnitBorder(Canvas canvas, int cid, int x, int y,
			int ulx, int uly, int base_x, int base_y,
			int lrx, int lry, int plateau_x, int plateau_y) {
		Point[] trapazoidList = new Point[4];
		int MAX_SIDES = 4;

		for (int i = 0; i < MAX_SIDES; i++) {
			trapazoidList[i] = new Point();
		}
		trapazoidList[0].x = x;
		trapazoidList[0].y = y;
		trapazoidList[1].x = base_x;
		trapazoidList[1].y = base_y;
		trapazoidList[2].x = lrx - base_x - ulx;
		trapazoidList[2].y = lry - base_y - uly;
		trapazoidList[3].x = -plateau_x;
		trapazoidList[3].y = -plateau_y;
		/*trapazoidList[4].x = ulx + plateau_x - lrx;
		trapazoidList[4].y = ulx + plateau_y - lry;*/
		for (int k = 0; k < 4; k++) {
			trapazoidList[k].x +=
				((k > 0) ? trapazoidList[k - 1].x : ulx);
			trapazoidList[k].y +=
				((k > 0) ? trapazoidList[k - 1].y : uly);
		}
		drawPolygon(canvas, cid, trapazoidList);
	}

	void drawUnit(Canvas canvas, boolean drawNext, int pmid, Colour cid,
			int x, int y, int i, int j, int width, int height) {
		int plain = (plainPiece) ? 1 : 0;
		int next = (drawNext) ? 1 : 0;

		if (canvas == null)
			return;
		Paint paint = new Paint();
		Rect rect = new Rect();
		paint.setColor(cid.getColor());
		rect.set(x + width * i, y + height * j,
			x + width * (i + 1), y + height * (j + 1));
		canvas.drawRect(rect, paint);
		Bitmap bitmap = pms[plain][next][pmid % NUM_BITMAPS];
		if (bitmap != null)
			canvas.drawBitmap(bitmap, null, rect, null);
	}

	void drawUnit(Canvas g, boolean drawNext, int pmid, int cid,
			int x, int y, int i, int j, int width, int height) {
		drawUnit(g, drawNext, pmid, colorThings[cid],
			x, y, i, j, width, height);
	}

	public static void clearUnits(Canvas canvas, Colour color,
			int x, int y, int i, int j, int width, int height) {
		Paint paint = new Paint();
		Rect rect = new Rect();
		paint.setColor(color.getColor());
		rect.set(x + width * i, y + height * j,
				x + width * (i + 1), y + height * (j + 1));
		canvas.drawRect(rect, paint);
	}

	public static void drawBorder(Canvas canvas,
			int x, int y, int width, int height) {
		Paint paint = new Paint();
		Rect rect = new Rect();
		paint.setColor(Colour.gray25);
		rect.set(0, 0, canvas.getWidth(), y);
		canvas.drawRect(rect, paint);
		rect.set(0, height, canvas.getWidth(), canvas.getHeight());
		canvas.drawRect(rect, paint);
		rect.set(0, y, x, height);
		canvas.drawRect(rect, paint);
		rect.set(width, y, canvas.getWidth(), height);
		canvas.drawRect(rect, paint);
	}

	void setPlainPiece(boolean plainPiece) {
		this.plainPiece = plainPiece;
	}

	boolean getPlainPiece() {
		return this.plainPiece;
	}
}
